; Dispatcher for Flapjax object storage.
(module flapjax-dispatcher mzscheme
  (require
   (lib "list.ss")
   (lib "etc.ss")
   (lib "contract.ss")
   (only "json.ss" json-read json-write)
   (lib "servlet.ss" "web-server")
   (lib "url.ss" "net")
   (lib "dispatch.ss" "web-server" "dispatchers")
   (prefix dispatcher: (lib "dispatch-lift.ss" "web-server" "dispatchers")))
  
  (define (errout . args)
    (apply fprintf (cons (current-error-port) args)))
  
  (define (report-invalid-request request)
    (errout "Invalid request: ~a~n" (url->string (request-uri request)))
    (json->string (format "invalid request sent to server: ~a" 
                          (url->string (request-uri request)))))
  
  (define-syntax (on-exn stx)
    (syntax-case stx ()
      [(_ handler-expr body-expr)
       (with-syntax ([exn (datum->syntax-object stx 'exn)])
         #`(with-handlers
               ([exn? (lambda (exn) handler-expr)])
             body-expr))]))
  
  ; last :: non-empty-list -> element
  (define (last lst)
    (cond
      [(empty? lst) (error 'last "given the empty list")]
      [(empty? (rest lst)) (first lst)]
      [else (last (rest lst))]))
  
  ; json->sexp :: json-string -> json-sexp
  (define (json->sexp json-string)
    (local ([define escaped-string-port (open-output-string)])
      (write json-string escaped-string-port)
      (local ([define escaped-string 
                (open-input-string (get-output-string escaped-string-port))])
        (json-read escaped-string))))
  
  ; json->string :: json->sexp -> json-string
  (define (json->string json-sexp)
    (local ([define out (open-output-string)])
      (parameterize ([current-output-port out])
        (on-exn
            (begin 
              (errout "Exception ~a (json-sexp: ~a)~n"
                      (exn-message exn) json-sexp)
              (raise exn))
          (json-write json-sexp)))
      (get-output-string out)))
  
  ; binds->json :: bindings -> json-sexp
  (define (binds->json binds)
    (list->vector 
     (map cons (map first binds) (map (λ (v) (json->sexp (cdr v))) binds))))
  
  ; make-obj-table :: -> (values get-dispatcher set-dispatcher)
  (provide/contract (make-obj-table (-> (values dispatcher? dispatcher?))))
  (define (make-obj-table)
    (local ([define ht (make-hash-table)]
            [define sema (make-semaphore 1)])
      (values (dispatcher:make (make-fx-responder ht sema obj-get))
              (dispatcher:make (make-fx-responder ht sema obj-set)))))
  
  ; make-list-table :: (values get-dispatcher cons-dispatcher)
  (provide/contract (make-list-table (-> (values dispatcher? dispatcher?))))
  (define (make-list-table)
    (local ([define ht (make-hash-table)]
            [define sema (make-semaphore 1)])
      (values (dispatcher:make (make-fx-responder ht sema list-get))
              (dispatcher:make (make-fx-responder ht sema list-cons)))))
  
  ; make-per-address-table :: (-> (values get-dispatcher set-dispatcher)
  ;                        -> (values get-dispatcher set-dispatcher)
  (provide/contract 
   (make-per-address-table 
    ((-> (values dispatcher? dispatcher?)) 
     . -> . (values dispatcher? dispatcher?))))
  (define (make-per-address-table make-get/set-dispatchers)
    ; TODO: The entries in this table should be removed on a timer.
    ; This should not be used to control Voyager III, for example.
    (local ([define ht (make-hash-table 'equal) #|strings as keys|#]
            [define sema (make-semaphore 1)]
            [define ((new-get/set-pair address))
              (let-values ([(get set) (make-get/set-dispatchers)])
                (hash-table-put! ht address (list get set))
                (list get set))]
            [define (get conn req)
              ((call-with-semaphore
                sema
                (λ ()
                  (first
                   (hash-table-get 
                    ht (request-client-ip req)
                    (new-get/set-pair (request-client-ip req))))))
               conn req)]
            [define (set conn req)
              ((call-with-semaphore
                sema
                (λ ()
                  (second 
                   (hash-table-get 
                    ht (request-client-ip req)
                    (new-get/set-pair (request-client-ip req))))))
               conn req)])
      (values get set)))
  
  ; make-fx-responder :: hashtable semaphore 
  ;                     (hashtable symbol request -> response)
  ;                   -> response
  (define ((make-fx-responder ht sema responder) request)
    (local ([define uri (request-uri request)]
            [define binds (request-bindings request)]
            [define path-elements (map path/param-path (url-path uri))])
      (if (>= (length path-elements) 3)
          (local ([define obj-name (string->symbol (last path-elements))])
            (call-with-semaphore 
             sema (λ () (responder ht obj-name request))))
          (report-invalid-request request))))
  
  (define (list-get ht name request)
    (local ([define list-value
              (hash-table-get ht name (λ () empty))])
      (json->string list-value)))
  
  (define (list-cons ht name request)
    (local ([define binds (request-bindings request)]
            [define obj-json (binds->json binds)])
      (hash-table-put! 
       ht name
       (cons obj-json (hash-table-get ht name (λ () empty))))
      (json->string "true" #| doesn't matter |#)))
  
  (define (obj-get ht name request)
    (local ([define binds (request-bindings request)]
            [define obj-value
              (hash-table-get ht name (λ () (binds->json binds)))])
      (json->string obj-value)))
  
  (define (obj-set ht name request)
    (local ([define binds (request-bindings request)]
            [define obj-json (binds->json binds)])
      (hash-table-put! ht name obj-json))
    (json->string "true") #| doesn't matter |#)
  
  ) ; end module